/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <errno.h>
#include <glib.h>
#include <sys/mman.h>
#include <tizen.h>
#include <dlog.h>
#include <wayland-client.h>
#include <Ecore_Wayland.h>
#include <Elementary.h>
#include <wayland-extension/tizen-extension-client-protocol.h>
#include <tizen-remote-surface-client-protocol.h>
#include <aul_screen_connector.h>

#include "screen_connector_provider.h"

#ifdef LOG_TAG
#undef LOG_TAG
#endif

#define LOG_TAG "SC_PROVIDER"

static int __init_count;
static struct tizen_remote_surface_manager *__rsm;

struct screen_connector_provider_s {
	struct wl_surface *surface;
	struct tizen_remote_surface_provider *rsp;
	char *id;
	Ecore_Wl_Window *win;
};

static void __rsp_resource_id_cb(void *data, struct tizen_remote_surface_provider *provider, uint32_t res_id)
{
	struct screen_connector_provider_s *remote = (struct screen_connector_provider_s *)data;

	if (remote == NULL) {
		LOGE("null remote");
		return;
	}

	LOGD("add app screen : %s, %d", remote->id, res_id);
	aul_screen_connector_add_app_screen(remote->id, res_id);
}

static void __rsp_visibility_cb(void *data, struct tizen_remote_surface_provider *provider, uint32_t visibility)
{
	struct screen_connector_provider_s *remote = (struct screen_connector_provider_s *)data;
	Ecore_Wl_Event_Window_Visibility_Change *ev;

	ev = calloc(1, sizeof(Ecore_Wl_Event_Window_Visibility_Change));
	if (!ev) {
		LOGE("out of memory");
		return;
	}

	ev->win = ecore_wl_window_id_get(remote->win);

	if (visibility == TIZEN_REMOTE_SURFACE_VISIBILITY_TYPE_VISIBLE)
		ev->fully_obscured = 0;
	else
		ev->fully_obscured = 1;

	LOGD("visibility changed:%d", visibility);

	ecore_event_add(ECORE_WL_EVENT_WINDOW_VISIBILITY_CHANGE, ev, NULL, NULL);
}

static const struct tizen_remote_surface_provider_listener __rsp_listener = {
	__rsp_resource_id_cb,
	__rsp_visibility_cb,
};

EXPORT_API screen_connector_provider_h screen_connector_provider_create(const char *id, struct wl_surface *surface)
{
	struct screen_connector_provider_s *remote = NULL;
	struct wl_display *display;
	struct wl_event_queue *queue = NULL;
	struct wl_display *rsm_wrapper = NULL;

	if (!__rsm) {
		LOGE("__rsm is not ready");
		goto out;
	}

	if (!surface || !id) {
		LOGE("invalid parameter");
		goto out;
	}

	rsm_wrapper = wl_proxy_create_wrapper(__rsm);
	if (!rsm_wrapper) {
		LOGE("failed to create wl display wrapper");
		goto out;
	}

	display = ecore_wl_display_get();
	if (!display) {
		LOGE("Fail to get");
		goto out;
	}

	queue = wl_display_create_queue(display);
	wl_proxy_set_queue((struct wl_proxy *)rsm_wrapper, queue);

	remote = (struct screen_connector_provider_s *)malloc(sizeof(struct screen_connector_provider_s));
	if (!remote) {
		LOGE("out of memory");
		goto out;
	}

	remote->surface = surface;
	remote->win = ecore_wl_window_surface_find(surface);
	if (!remote->win) {
		LOGE("failed to find win");
		free(remote);
		remote = NULL;
		goto out;
	}

	remote->rsp = tizen_remote_surface_manager_create_provider((struct tizen_remote_surface_manager *)rsm_wrapper, surface);
	if (!remote->rsp) {
		LOGE("failed to create provider");
		free(remote);
		remote = NULL;
		goto out;
	}

	remote->id = strdup(id);
	if (!remote->id) {
		LOGE("out of memory");
		free(remote);
		remote = NULL;
		goto out;
	}

	tizen_remote_surface_provider_add_listener(remote->rsp, &__rsp_listener, remote);
	wl_display_roundtrip_queue(display, queue);
	wl_proxy_set_queue((struct wl_proxy *)remote->rsp, NULL);
	LOGD("surface remote enabled");

out:
	if (queue)
		wl_event_queue_destroy(queue);

	if (rsm_wrapper)
		wl_proxy_wrapper_destroy(rsm_wrapper);

	return remote;
}

EXPORT_API int screen_connector_provider_remote_enable(const char *id, struct wl_surface *surface)
{
	screen_connector_provider_h handle;

	handle = screen_connector_provider_create(id, surface);
	if (handle)
		return 0;

	return -1;
}

EXPORT_API int screen_connector_provider_set_event_filter(screen_connector_provider_h provider, unsigned int filter)
{
	if (!provider) {
		LOGE("invalid arguments");
		return -1;
	}

	if (tizen_remote_surface_provider_get_version(provider->rsp) <
		TIZEN_REMOTE_SURFACE_PROVIDER_SET_INPUT_EVENT_FILTER_SINCE_VERSION) {
		LOGE("unsupported operation");
		return -1;
	}

	tizen_remote_surface_provider_set_input_event_filter(provider->rsp, filter);

	return 0;
}

EXPORT_API int screen_connector_provider_init(void)
{
	struct wl_registry *registry;
	Ecore_Wl_Global *global;
	Eina_Inlist *globals;

	if (__init_count > 0)
		return 0;

	ecore_wl_init(NULL);

	registry = ecore_wl_registry_get();
	globals = ecore_wl_globals_get();

	if (!registry || !globals) {
		LOGE("could not get registry(%p) or global list(%p)", registry,
				globals);
		return -1;
	}

	EINA_INLIST_FOREACH(globals, global) {
		if (!strcmp(global->interface, "tizen_remote_surface_manager")) {
			__rsm = wl_registry_bind(registry, global->id,
				&tizen_remote_surface_manager_interface,
				global->version < 4 ? global->version : 4);
		}
	}

	if (!__rsm) {
		LOGE("could not get remote surface manager");
		return -1;
	}
	__init_count++;

	return 0;
}

EXPORT_API int screen_connector_provider_fini(void)
{
	if (__init_count > 1)
		return 0;

	if (__rsm)
		tizen_remote_surface_manager_destroy(__rsm);

	ecore_wl_shutdown();

	__init_count--;
	return 0;
}
